% Instantiates all ramifications.


% =============================================================================================
% Finds the relevant Literals, given Delta.
% =============================================================================================

find_relevant_literals_from_Delta(Delta, RelLiterals, RelRams, RelActions):-
	convert_Delta_to_Literals_at_T(Delta, T, DeltaLiterals),
	find_relevant_literals(DeltaLiterals, RelLiterals, RelRams, RelActions).


% =============================================================================================
% Finds the relevant Literals, given an initial set of Literals.
% =============================================================================================

find_relevant_literals(Literals, RelLiterals, RelevantRams, RelevantActions):-
	find_all_rams(Rams),
	find_all_actions(Actions),
	find_all_relevant(Literals, Rams, Actions, RelevantRams, RelevantActions),
	convert_Rams_to_Literals(RelevantRams, RelLiteralsFromRams),
	convert_Actions_to_Literals(RelevantActions, RelLiteralsFromActions),
	myappend(Literals, RelLiteralsFromRams, RelLiterals1),
	myappend(RelLiterals1, RelLiteralsFromActions, RelLiterals2),
	removeduplicates(RelLiterals2, RelLiterals).


find_all_relevant(Literals, Rams, Actions, RelevantRams, RelevantActions):-
	find_relevant_rams(Literals, Rams, OneRams),
	intersection(Rams, OneRams, [_|_]), !, 
	convert_Rams_to_Literals(OneRams, RamsLiterals),
   	union(RamsLiterals, Literals, NewLiterals), 
   	subtract(Rams, OneRams, Rams1), 
	find_all_relevant(NewLiterals, Rams1, Actions, RestRams, RelevantActions),
	myappend(OneRams, RestRams, RelevantRams).
find_all_relevant(Literals, Rams, Actions, RelevantRams, RelevantActions):-
	find_relevant_actions(Literals, Actions, OneActions),
	intersection(Actions, OneActions, [_|_]), !, 
	convert_Actions_to_Literals(OneActions, ActionsLiterals),
   	union(ActionsLiterals, Literals, NewLiterals), 
   	subtract(Actions, OneActions, Actions1), 
	find_all_relevant(NewLiterals, Rams, Actions1, RelevantRams, RestActions),
	myappend(OneActions, RestActions, RelevantActions).
find_all_relevant(_Literals, _Rams, _Actions, [], []).


% Relevant Rams
% =============

find_all_rams(Rams):-
	findall(ramification(Y,GroundListRest), (clause(ram(Y),Rest),convert(Rest,ListRest),ground(ListRest,GroundListRest)), Rams).

find_relevant_rams(_Literals, [], []).
find_relevant_rams(Literals, [ramification(H,B)|Rams], [ramification(H,B)|OneRams]):-
	mymember(H, Literals), !,
	find_relevant_rams(Literals, Rams, OneRams).
find_relevant_rams(Literals, [ramification(H,B)|Rams], [ramification(H,B)|OneRams]):-
	contrary(H, NH), mymember(NH, Literals), !,
	find_relevant_rams(Literals, Rams, OneRams).
find_relevant_rams(Literals, [ramification(H,B)|Rams], [ramification(H,B)|OneRams]):-
   	intersection(Literals, B, [_|_]), !,
	find_relevant_rams(Literals, Rams, OneRams).
find_relevant_rams(Literals, [ramification(H,B)|Rams], [ramification(H,B)|OneRams]):-
	contrary_all(B, NB), intersection(Literals, NB, [_|_]), !,
	find_relevant_rams(Literals, Rams, OneRams).
find_relevant_rams(Literals, [_|Rams], OneRams):-
	find_relevant_rams(Literals, Rams, OneRams).

convert_Rams_to_Literals([], []).
convert_Rams_to_Literals([ramification(H,B)|Rams], RamsLiterals):-
	convert_Rams_to_Literals(Rams, RamsLiterals1),
	myappend(RamsLiterals1, [H|B], RamsLiterals).


% Relevant Actions
% ================

find_all_actions(Actions):-
	findall(action(A,holds(Y,T),GroundListRest), (clause(initiation(Y,T),(A,Rest)),convert(Rest,ListRest),ground(ListRest,GroundListRest)), Actions1),
	findall(action(A,neg(holds(Y,T)),GroundListRest), (clause(termination(Y,T),(A,Rest)),convert(Rest,ListRest),ground(ListRest,GroundListRest)), Actions2),
	myappend(Actions1, Actions2, Actions).

find_relevant_actions(_Literals, [], []).
find_relevant_actions(Literals, [action(A,H,B)|Actions], [action(A,H,B)|OneActions]):-
	mymember(H, Literals), action_happens(A), !,
	find_relevant_actions(Literals, Actions, OneActions).
find_relevant_actions(Literals, [action(A,H,B)|Actions], [action(A,H,B)|OneActions]):-
	contrary(H, NH), mymember(NH, Literals), action_happens(A), !,
	find_relevant_actions(Literals, Actions, OneActions).
find_relevant_actions(Literals, [action(A,H,B)|Actions], [action(A,H,B)|OneActions]):-
   	intersection(Literals, B, [_|_]), action_happens(A), !,
	find_relevant_actions(Literals, Actions, OneActions).
find_relevant_actions(Literals, [action(A,H,B)|Actions], [action(A,H,B)|OneActions]):-
	contrary_all(B, NB), intersection(Literals, NB, [_|_]), action_happens(A), !,
	find_relevant_actions(Literals, Actions, OneActions).
find_relevant_actions(Literals, [_|Actions], OneActions):-
	find_relevant_actions(Literals, Actions, OneActions).

convert_Actions_to_Literals([], []).
convert_Actions_to_Literals([action(_A,H,B)|Actions], ActionsLiterals):-
	convert_Actions_to_Literals(Actions, ActionsLiterals1),
	myappend(ActionsLiterals1, [H|B], ActionsLiterals).


% =============================================================================================
% Finds the relevant T-propositions, given an initial set of Literals.
% =============================================================================================

find_relevant_tprops(Literals, RelTprops):-
	find_all_tprops(Tprops),
	find_all_relevant_tprops(Literals, Tprops, RelTprops).


find_all_tprops(Tprops):-
	findall(Y, (clause(tprop(Y),true), get_fluent(Y,F), \+constantfluent(F)), Tprops1),
	find_relevant_pprop_literals(RelPpropLiterals),
	myappend(Tprops1, RelPpropLiterals, Tprops2),
	removeduplicates(Tprops2, Tprops).


find_all_relevant_tprops(_Literals, [], []).
find_all_relevant_tprops(Literals, [Y|Tprops], [Y|RelevantTprops]):-
	settime(T, Y, TY), % Protect time variables from unification (singleton variable).
	mymember(TY, Literals), !,
	find_all_relevant_tprops(Literals, Tprops, RelevantTprops).
find_all_relevant_tprops(Literals, [Y|Tprops], [Y|RelevantTprops]):-
	settime(T, Y, TY), % Protect time variables from unification (singleton variable).
	contrary(TY, NY), mymember(NY, Literals), !,
	find_all_relevant_tprops(Literals, Tprops, RelevantTprops).
find_all_relevant_tprops(Literals, [_|Tprops], RelevantTprops):-
	find_all_relevant_tprops(Literals, Tprops, RelevantTprops).


% =============================================================================================
% Finds the relevant P-proposition literals.
% =============================================================================================

find_relevant_pprop_literals(RelPpropLiterals):-
	findall(GroundedPrec,(clause(pprop(A,T,Prec),true), clause(happens(A,T),true), ground(Prec,GroundedPrec)), RelPpropLists),
	list_append(RelPpropLists, RelPpropLiterals).


% =============================================================================================
% Finds the relevant Time Points, given the relevant Actions.
% =============================================================================================

find_relevant_time_points(RelevantActions, RelevantTimes):-
	findall(NT,(clause(happens(L,T),true), mymember(action(happens(L,T), _, _), RelevantActions), nexttime(T, NT)), Times),
%	findall(NT,(clause(happens(_,T),true), nexttime(T, NT)), Times),
	addzero(Times,TimesZero),
   	removeduplicates(TimesZero, CTimes),
   	order(CTimes,RelevantTimes).


% =============================================================================================
% End of file.
% =============================================================================================
